import zlib


import numpy as np
import struct


from .abc import Codec
from .compat import ensure_contiguous_ndarray, ndarray_copy
from .jenkins import jenkins_lookup3


class Checksum32(Codec):

    # override in sub-class
    checksum = None

    def encode(self, buf):
        arr = ensure_contiguous_ndarray(buf).view('u1')
        checksum = self.checksum(arr) & 0xffffffff
        enc = np.empty(arr.nbytes + 4, dtype='u1')
        enc[:4].view('<u4')[0] = checksum
        ndarray_copy(arr, enc[4:])
        return enc

    def decode(self, buf, out=None):
        arr = ensure_contiguous_ndarray(buf).view('u1')
        expect = arr[:4].view('<u4')[0]
        checksum = self.checksum(arr[4:]) & 0xffffffff
        if expect != checksum:
            raise RuntimeError('checksum failed')
        return ndarray_copy(arr[4:], out)


class CRC32(Checksum32):

    codec_id = 'crc32'
    checksum = zlib.crc32


class Adler32(Checksum32):

    codec_id = 'adler32'
    checksum = zlib.adler32


class JenkinsLookup3(Checksum32):
    """Bob Jenkin's lookup3 checksum with 32-bit output

    This is the HDF5 implementation.
    https://github.com/HDFGroup/hdf5/blob/577c192518598c7e2945683655feffcdbdf5a91b/src/H5checksum.c#L378-L472

    With this codec, the checksum is concatenated on the end of the data
    bytes when encoded. At decode time, the checksum is performed on
    the data portion and compared with the four-byte checksum, raising
    RuntimeError if inconsistent.

    Attributes:
       initval: initial seed passed to the hash algorithm, default: 0
       prefix: bytes prepended to the buffer before evaluating the hash, default: None
    """

    checksum = jenkins_lookup3
    codec_id = "jenkins_lookup3"

    def __init__(self, initval: int = 0, prefix=None):
        self.initval = initval
        if prefix is None:
            self.prefix = None
        else:
            self.prefix = np.frombuffer(prefix, dtype='uint8')

    def encode(self, buf):
        """Return buffer plus 4-byte Bob Jenkin's lookup3 checksum"""
        buf = ensure_contiguous_ndarray(buf).ravel().view('uint8')
        if self.prefix is None:
            val = jenkins_lookup3(buf, self.initval)
        else:
            val = jenkins_lookup3(np.hstack((self.prefix, buf)), self.initval)
        return buf.tobytes() + struct.pack("<I", val)

    def decode(self, buf, out=None):
        """Check Bob Jenkin's lookup3 checksum, and return buffer without it"""
        b = ensure_contiguous_ndarray(buf).view('uint8')
        if self.prefix is None:
            val = jenkins_lookup3(b[:-4], self.initval)
        else:
            val = jenkins_lookup3(np.hstack((self.prefix, b[:-4])), self.initval)
        found = b[-4:].view("<u4")[0]
        if val != found:
            raise RuntimeError(
                f"The Bob Jenkin's lookup3 checksum of the data ({val}) did not"
                f" match the expected checksum ({found}).\n"
                "This could be a sign that the data has been corrupted."
            )
        if out is not None:
            out.view("uint8")[:] = b[:-4]
            return out
        return memoryview(b[:-4])



import numpy as np
import pytest

from numcodecs.jenkins import jenkins_lookup3
from numcodecs.checksum32 import JenkinsLookup3


def test_jenkins_lookup3():
    h = jenkins_lookup3(b"", 0)
    assert h == 0xdeadbeef
    h = jenkins_lookup3(b"", 0xdeadbeef)
    assert h == 0xbd5b7dde
    h = jenkins_lookup3(b"Four score and seven years ago", 0)
    assert h == 0x17770551
    h = jenkins_lookup3(b"Four score and seven years ago", 1)
    assert h == 0xcd628161

    # jenkins-cffi example
    h = jenkins_lookup3(b"jenkins", 0)
    assert h == 202276345

    h_last = [0]

    h = jenkins_lookup3(b"", h_last[-1])
    assert h not in h_last
    h_last.append(h)

    h = jenkins_lookup3(b"", h_last[-1])
    assert h not in h_last
    h_last.append(h)

    h = jenkins_lookup3(b"", h_last[-1])
    assert h not in h_last
    h_last.append(h)

    h = jenkins_lookup3(b"", h_last[-1])
    assert h not in h_last
    h_last.append(h)

    h = jenkins_lookup3(b"", h_last[-1])
    assert h not in h_last
    h_last.append(h)

    h = jenkins_lookup3(b"", h_last[-1])
    assert h not in h_last
    h_last.append(h)

    h = jenkins_lookup3(b"", h_last[-1])
    assert h not in h_last
    h_last.append(h)

    h = jenkins_lookup3(b"", h_last[-1])
    assert h not in h_last
    h_last.append(h)

    h = jenkins_lookup3(b"", h_last[-1])
    assert h not in h_last
    h_last.append(h)

    a = np.frombuffer(b"Four score and seven years ago", dtype="uint8")
    h = jenkins_lookup3(a, 0)
    assert h == 0x17770551


def test_jenkins_lookup3_codec():
    s = b"Four score and seven years ago"
    j = JenkinsLookup3()
    result = j.encode(s)
    assert result[-4:] == b'\x51\x05\x77\x17'
    assert bytes(j.decode(result)) == s

    j = JenkinsLookup3(initval=0xdeadbeef)
    result = j.encode(s)
    assert bytes(j.decode(result)) == s

    j = JenkinsLookup3(initval=1230)
    result = j.encode(s)
    assert result[-4:] == b'\xd7Z\xe2\x0e'
    assert bytes(j.decode(result)) == s

    j = JenkinsLookup3(initval=1230, prefix=b"Hello world")
    result = j.encode(s)
    assert bytes(j.decode(result)) == s

    chunk_index = b"\x00\x08\x00\x00\x00\x00\x00\x00" + \
        b"\xf7\x0f\x00\x00\x00\x00\x00\x00" + \
        b"\xf7\x17\x00\x00\x00\x00\x00\x00" + \
        b"\xf7\x0f\x00\x00\x00\x00\x00\x00" + \
        b"\xee'\x00\x00\x00\x00\x00\x00" + \
        b"\xf7\x0f\x00\x00\x00\x00\x00\x00" + \
        b"\xe57\x00\x00\x00\x00\x00\x00" + \
        b"\xf7\x0f\x00\x00\x00\x00\x00\x00" + \
        b"\xdcG\x00\x00\x00\x00\x00\x00" + \
        b"\xf7\x0f\x00\x00\x00\x00\x00\x00" + \
        b"\xd3W\x00\x00\x00\x00\x00\x00" + \
        b"\xf7\x0f\x00\x00\x00\x00\x00\x00" + \
        b"\xcag\x00\x00\x00\x00\x00\x00" + \
        b"\xf7\x0f\x00\x00\x00\x00\x00\x00" + \
        b"\xc1w\x00\x00\x00\x00\x00\x00" + \
        b"\xf7\x0f\x00\x00\x00\x00\x00\x00" + \
        b"\xb8\x87\x00\x00\x00\x00\x00\x00" + \
        b"\xf7\x0f\x00\x00\x00\x00\x00\x00" + \
        b"\xaf\x97\x00\x00\x00\x00\x00\x00" + \
        b"\xf7\x0f\x00\x00\x00\x00\x00\x00" + \
        b"\xa6\xa7\x00\x00\x00\x00\x00\x00" + \
        b"\xf7\x0f\x00\x00\x00\x00\x00\x00" + \
        b"\x9d\xb7\x00\x00\x00\x00\x00\x00" + \
        b"\xf7\x0f\x00\x00\x00\x00\x00\x00" + \
        b"\x94\xc7\x00\x00\x00\x00\x00\x00" + \
        b"\xf7\x0f\x00\x00\x00\x00\x00\x00" + \
        b"\x8b\xd7\x00\x00\x00\x00\x00\x00" + \
        b"\xf7\x0f\x00\x00\x00\x00\x00\x00" + \
        b"\x82\xe7\x00\x00\x00\x00\x00\x00" + \
        b"\xf7\x0f\x00\x00\x00\x00\x00\x00" + \
        b"y\xf7\x00\x00\x00\x00\x00\x00" + \
        b"\xf7\x0f\x00\x00\x00\x00\x00\x00" + \
        b"n\x96\x07\x85"
    hdf5_fadb_prefix = b'FADB\x00\x01\xcf\x01\x00\x00\x00\x00\x00\x00'
    j = JenkinsLookup3(prefix=hdf5_fadb_prefix)
    result = j.encode(chunk_index[:-4])
    j.decode(result)
    assert result == chunk_index


@pytest.mark.parametrize(
    "dtype",
    ["uint8", "int32", "float32"]
)
def test_with_data(dtype):
    data = np.arange(100, dtype=dtype)
    j = JenkinsLookup3()
    arr = np.frombuffer(j.decode(j.encode(data)), dtype=dtype)
    assert (arr == data).all()


def test_error():
    data = np.arange(100)
    j = JenkinsLookup3()
    enc = j.encode(data)
    enc2 = bytearray(enc)
    enc2[0] += 1
    with pytest.raises(RuntimeError) as e:
        j.decode(enc2)
    assert "Bob Jenkin's lookup3 checksum" in str(e.value)


def test_out():
    data = np.frombuffer(bytearray(b"Hello World"), dtype="uint8")
    j = JenkinsLookup3()
    result = j.encode(data)
    j.decode(result, out=data)



import numpy as np
import pytest

from numcodecs.fletcher32 import Fletcher32


@pytest.mark.parametrize(
    "dtype",
    ["uint8", "int32", "float32"]
)
def test_with_data(dtype):
    data = np.arange(100, dtype=dtype)
    f = Fletcher32()
    arr = np.frombuffer(f.decode(f.encode(data)), dtype=dtype)
    assert (arr == data).all()


def test_error():
    data = np.arange(100)
    f = Fletcher32()
    enc = f.encode(data)
    enc2 = bytearray(enc)
    enc2[0] += 1
    with pytest.raises(RuntimeError) as e:
        f.decode(enc2)
    assert "fletcher32 checksum" in str(e.value)


def test_known():
    data = (
        b'w\x07\x00\x00\x00\x00\x00\x00\x85\xf6\xff\xff\xff\xff\xff\xff'
        b'i\x07\x00\x00\x00\x00\x00\x00\x94\xf6\xff\xff\xff\xff\xff\xff'
        b'\x88\t\x00\x00\x00\x00\x00\x00i\x03\x00\x00\x00\x00\x00\x00'
        b'\x93\xfd\xff\xff\xff\xff\xff\xff\xc3\xfc\xff\xff\xff\xff\xff\xff'
        b"'\x02\x00\x00\x00\x00\x00\x00\xba\xf7\xff\xff\xff\xff\xff\xff"
        b'\xfd%\x86d')
    data3 = Fletcher32().decode(data)
    outarr = np.frombuffer(data3, dtype="<i8")
    expected = [
        1911, -2427, 1897, -2412, 2440, 873, -621, -829, 551, -2118,
    ]
    assert outarr.tolist() == expected


def test_out():
    data = np.frombuffer(bytearray(b"Hello World"), dtype="uint8")
    f = Fletcher32()
    result = f.encode(data)
    f.decode(result, out=data)



import os.path
import pkgutil
import sys

import pytest

from multiprocessing import Process

import numcodecs.registry

if not pkgutil.find_loader("importlib_metadata"):  # pragma: no cover
    pytest.skip("This test module requires importlib_metadata to be installed")

here = os.path.abspath(os.path.dirname(__file__))


def get_entrypoints_with_importlib_metadata_loaded():
    # importlib_metadata patches importlib.metadata, which can lead to breaking changes
    # to the APIs of EntryPoint objects used when registering entrypoints. Attempt to
    # isolate those changes to just this test.
    import importlib_metadata  # noqa: F401
    sys.path.append(here)
    numcodecs.registry.run_entrypoints()
    cls = numcodecs.registry.get_codec({"id": "test"})
    assert cls.codec_id == "test"


def test_entrypoint_codec_with_importlib_metadata():
    p = Process(target=get_entrypoints_with_importlib_metadata_loaded)
    p.start()
    p.join()
    assert p.exitcode == 0



from numcodecs.abc import Codec


class TestCodec(Codec):

    codec_id = "test"

    def encode(self, buf):  # pragma: no cover
        pass

    def decode(self, buf, out=None):  # pragma: no cover
        pass
